/*
 *   $Id: defmgr.c,v 2.5 2000/09/26 16:44:37 ondra Exp $
 * 
 *   Scientific Calculator for Palm's and PilmPilot's.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include "konvert.h"
#include "calcDB.h"
#include "defmgr.h"
#include "display.h"
#include "stack.h"
#include "calc.h"
#include "calcrsc.h"
#include "varmgr.h"

dbList *varItems=NULL;
static char HORIZ_ELLIPSIS;

#define VARNAME_WIDTH  40

/***********************************************************************
 *
 * FUNCTION:     def_draw_item
 * 
 * DESCRIPTION:  Custom drawing function for list, it draws the line
 *               as a table of 2 item, separated by space,
 *               computes items on-the-fly, because it is faster
 *               then allocating the whole thing in advance 
 *               and consumes less memory
 *
 * PARAMETERS:   read PalmOS Reference
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void def_draw_item(Int16 itemNum,RectangleType *bounds,
						  Char **itemText) IFACE;
static void
def_draw_item(Int16 itemNum,RectangleType *bounds,
			  Char **itemText)
{
	Int16 maxwidth,len;
	char *text;	
	CError err;
	Trpn item;
	
	if (varItems->type[itemNum]==function) {
		text = MemPtrNew(MAX_FUNCNAME+10);
		StrCopy(text,itemText[itemNum]);
		StrCat(text,"()");
		WinDrawChars(text,StrLen(text),bounds->topLeft.x,bounds->topLeft.y);
		MemPtrFree(text);
		
		err = db_func_description(varItems->list[itemNum],&text);
		if (err) 
		  text = print_error(err);
	}
	else {
		WinDrawChars(itemText[itemNum],StrLen(itemText[itemNum]),
					 bounds->topLeft.x,bounds->topLeft.y);
		
		item = db_read_variable(varItems->list[itemNum],&err);
		if (!err) {
			text = display_default(item);
			rpn_delete(item);
		}
		else
		  text = print_error(err);
	}
	
	maxwidth = bounds->extent.x - VARNAME_WIDTH-3;
	len = StrLen(text);
	/* Umm.. FntDrawTruncChars is a PalmOS3.1 feature :-( */
	if (FntCharsWidth(text,len)>maxwidth) {
		while (FntCharsWidth(text,len)>maxwidth)
		  len--;
		text[len-1]=HORIZ_ELLIPSIS;
	}	
	WinDrawChars(text,len,bounds->topLeft.x+VARNAME_WIDTH+3,
				 bounds->topLeft.y);
	MemPtrFree(text);
}

/***********************************************************************
 *
 * FUNCTION:     def_init_varlist
 * 
 * DESCRIPTION:  Initialize the on-screen list with names of variables
 *               and their values
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
def_init_varlist(void) IFACE;
static void
def_init_varlist(void)
{
	ListPtr list;
	
	varItems = db_get_list(notype);
	if (!varItems->size) {
		db_delete_list(varItems);
		varItems = NULL;
		LstSetListChoices(GetObjectPtr(defList),NULL,0);
		return;
	}
	list = GetObjectPtr(defList);
	LstSetListChoices(list,varItems->list,varItems->size);
	LstSetDrawFunction(list,def_draw_item);
}

/***********************************************************************
 *
 * FUNCTION:     def_destroy_varlist
 * 
 * DESCRIPTION:  Free dynamically allocated structures for
 *               the onscreen field
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
def_destroy_varlist(void) IFACE;
static void
def_destroy_varlist(void)
{
	if (!varItems)
	  return;
	
	db_delete_list(varItems);
	varItems = NULL;
}

static void
defmgr_update_callback(void) IFACE;
static void
defmgr_update_callback(void)
{
	FrmUpdateForm(defForm,frmUpdateVars);
}

/***********************************************************************
 *
 * FUNCTION:     DefmgrHandleEvent
 * 
 * DESCRIPTION:  Event handler for Definition manager main window
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean 
DefmgrHandleEvent(EventPtr event)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 controlID;
	Int16 i;

	switch (event->eType) {
	 case keyDownEvent: {
		 ListPtr list;
		 
		 frm = FrmGetActiveForm();
		 list = GetObjectPtr(defList);
		 switch (event->data.keyDown.chr) {
		  case pageUpChr: 
			 FrmHideObject(frm,FrmGetObjectIndex(frm,defList));
			 LstScrollList(list,winUp,LstGetVisibleItems(list)-1);
			 FrmShowObject(frm,FrmGetObjectIndex(frm,defList));
			 handled = true;
			 break;
		  case pageDownChr:
			 FrmHideObject(frm,FrmGetObjectIndex(frm,defList));
			 LstScrollList(list,winDown,LstGetVisibleItems(list)-1);
			 FrmShowObject(frm,FrmGetObjectIndex(frm,defList));
			 handled = true;
			 break;
		 }
		 break;
	 }
	 case ctlSelectEvent:
		controlID = event->data.ctlSelect.controlID;
		switch (controlID) {
		 case defDone:
			def_destroy_varlist();
			FrmReturnToForm(0);
			handled = true;
			break;
		 case defModify:
			i = LstGetSelection(GetObjectPtr(defList));
			if (i != noListSelection) {
				if (varItems->type[i] == function)
				  varmgr_edit(varItems->list[i],"Function Modification",
							  function,defmgr_update_callback,false);
				else
				  varmgr_edit(varItems->list[i],"Variable Modification",
							  variable,defmgr_update_callback,false);
			}
			handled = true;
			break;
		 case defNew:
			i = LstPopupList(GetObjectPtr(defPopupList));
			if (i==0) /* User has choosen 'function' */
			  varmgr_edit("","New Function",function,
						  defmgr_update_callback,true);
			else if (i==1) /* User has choosen 'variable' */
			  varmgr_edit("","New Variable",variable,
						  defmgr_update_callback,true);
			  
			handled = true;
			break;
		 case defDelete:
			handled = true;
			i = LstGetSelection(GetObjectPtr(defList));
			
			if (i == noListSelection)
			  break;
			if (!FrmCustomAlert(defConfirmDelete,
								varItems->type[i]==variable?"variable":"function",
								varItems->list[i],
								NULL)) {
				db_delete_record(varItems->list[i]);
				def_destroy_varlist();
				def_init_varlist();
				LstDrawList(GetObjectPtr(defList));
			}
			break;
		}
		break;
	 case frmOpenEvent:
		ChrHorizEllipsis(&HORIZ_ELLIPSIS);
		frm = FrmGetActiveForm();
		
		def_init_varlist();
		
		FrmDrawForm(frm);
		
		handled = true;
		break;		
	 case frmCloseEvent:
		def_destroy_varlist();
		/* FrmHandleEvent must handle this event itself */
		handled = false;
		break;
	 case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateVars) {
			def_destroy_varlist();
			def_init_varlist();
			LstDrawList(GetObjectPtr(defList));
			handled = true;
		}
		break;
	}
	
	return handled;
}
