/*
 *   $Id: graph.c,v 2.33 2000/10/01 22:25:38 ondra Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/


#include <PalmOS.h>

#include "defuns.h"
#include "MathLib.h"
#include "konvert.h"
#include "funcs.h"
#include "calcDB.h"
#include "calcrsc.h"
#include "prefs.h"
#include "calc.h"
#include "graph.h"
#include "grprefs.h"
#include "stack.h"
#include "grsetup.h"
#include "grtaps.h"

static struct {
	Boolean active;
	Int32 position;
	Int16 funcnum;
	double start;
	double stop;
	double step;
}graphState;

/* Screen preferences in pixels */
/* Please note, tha ymax is de facto smaller then ymin on the screen */
struct {
	/* The Y screen max is the smallest value */
	/*0,0   1,0                                     */
	/*0,1                                        */
	/*0,2[beginning of graph                     */
	 
	Coord xmin,ymax;
	Coord width,height;
	/* Cached coeficients for graph2screen computations */
	/* CacheTrans MUST be called when any of the ScrPrefs
	 * and graphPrefs attributes is changed */
	double xtrans,ytrans;
}ScrPrefs;

Tgraph_curves graphCurves[MAX_GRFUNCS];

/***********************************************************************
 *
 * FUNCTION:     graph_is_onscreen
 * 
 * DESCRIPTION:  If the coordinates fit on the screen
 *
 * PARAMETERS:   x,y - coordinates of the original function
 *
 * RETURN:       true - fits, false - outside
 *      
 ***********************************************************************/
static Boolean
graph_is_onscreen(double x,double y)
{
	if (!finite(x) || !finite(y))
	  return false;
	if (x>=graphPrefs.xmin && x<=graphPrefs.xmax &&
		y>=graphPrefs.ymin && y<=graphPrefs.ymax)
	  return true;
	return false;
}

/***********************************************************************
 *
 * FUNCTION:     graph_xgr2scr
 * 
 * DESCRIPTION:  Convert the 'x' or 'y' coordinate from the function to
 *               screen coordinates
 * 
 * PARAMETERS:   x or y
 *
 * RETURN:       coordinate
 *      
 ***********************************************************************/
Coord 
graph_xgr2scr(double x)
{
	return ScrPrefs.xmin+(Coord)round((x-graphPrefs.xmin)*ScrPrefs.xtrans);
}

Coord 
graph_ygr2scr(double y)
{
	return ScrPrefs.ymax+(Coord)round((graphPrefs.ymax-y)*ScrPrefs.ytrans);
}

/***********************************************************************
 *
 * FUNCTION:     graph_xscr2gr
 * 
 * DESCRIPTION:  Convert a screen 'x' (or y) coordinate to a graph coordinate
 *
 * PARAMETERS:   x or y
 *
 * RETURN:       coordinate
 *      
 ***********************************************************************/
double 
graph_xscr2gr(Coord x)
{
	return graphPrefs.xmin+(double)(x-ScrPrefs.xmin)/(double)ScrPrefs.xtrans;
}

double 
graph_yscr2gr(Coord y)
{
	return graphPrefs.ymin+(double)((ScrPrefs.ymax+ScrPrefs.height)-y)/ScrPrefs.ytrans;
//	return graphPrefs.ymax-(double)(y-ScrPrefs.ymax)/(double)ScrPrefs.ytrans;
}

/***********************************************************************
 *
 * FUNCTION:     graph_isec_y
 * 
 * DESCRIPTION:  Find an intersection of a line with a vertical line 
 *               - it is used when drawing lines to points outside of
 *               the drawing are
 *
 * PARAMETERS:   x1,y1 - point 1
 *               x2,y2 - point 2
 *               x3 - the vertical line
 * RETURN:       y3 - the point on the vertical line
 *      
 ***********************************************************************/
static double
graph_isec_y(double x1,double y1,double x2,double y2,double x3)
{
	double sidea,sideb,parta;
	
	sidea = y2-y1;
	sideb = x2-x1;
	parta = x2 - x3;
	
	return y2-(sidea*parta)/sideb;
}

#define graph_isec_x(x1,y1,x2,y2,y3) (graph_isec_y(y1,x1,y2,x2,y3))

/***********************************************************************
 *
 * FUNCTION:     graph_outscreen
 * 
 * DESCRIPTION:  Find a point, where a line should be drawn instead of
 *               the point outside of drawing are
 *
 * PARAMETERS:   onx,ony - points in the drawing area
 *               offx,offy - points outside drawing area
 *
 * RETURN:       newx,newy - new points on the border of drawing area
 *      
 ***********************************************************************/
static void
graph_outscreen(double *newx,double *newy,
				double onx,double ony,double offx,double offy)
{
	double tmp;

	tmp = graph_isec_x(onx,ony,offx,offy,graphPrefs.ymin);
	if (offy<graphPrefs.ymin && tmp>=graphPrefs.xmin && tmp<=graphPrefs.xmax) {
		*newx = tmp;
		*newy = graphPrefs.ymin;
		return;
	}
	tmp = graph_isec_x(onx,ony,offx,offy,graphPrefs.ymax);
	if (offy>graphPrefs.ymax && tmp>=graphPrefs.xmin && tmp<=graphPrefs.xmax) {
		*newx = tmp;
		*newy = graphPrefs.ymax;
		return;
	} 
	
	tmp = graph_isec_y(onx,ony,offx,offy,graphPrefs.xmin);
	if (offx<graphPrefs.xmin && tmp>=graphPrefs.ymin && tmp<=graphPrefs.ymax) {
		*newx = graphPrefs.xmin;
		*newy = tmp;
		return;
	}
	tmp = graph_isec_y(onx,ony,offx,offy,graphPrefs.xmax);
	if (offx>graphPrefs.xmax && tmp>=graphPrefs.ymin && tmp<=graphPrefs.ymax) {
		*newx = graphPrefs.xmax;
		*newy = tmp;
		return;
	} 
	*newx = onx;
	*newy = ony;
}

/***********************************************************************
 *
 * FUNCTION:     graph_new_point
 * 
 * DESCRIPTION:  Adds a new point to graph. Depending on the selected
 *               style this function chooses way of drawing, colors etc.
 *
 * PARAMETERS:   x,y - coordinates of a new point
 *               funcnum - function number, where the point belongs to
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_new_point(double x,double y,Int16 funcnum)
{
	double oldx = graphCurves[funcnum].x;
	double oldy = graphCurves[funcnum].y;
		
	if (!graphCurves[funcnum].valid &&
		graph_is_onscreen(x,y)) {
		WinDrawLine(graph_xgr2scr(x),graph_ygr2scr(y),
					graph_xgr2scr(x),graph_ygr2scr(y));
	}	
	else if (graphCurves[funcnum].valid) {
		if (graph_is_onscreen(x,y) && graph_is_onscreen(oldx,oldy)) { 
			WinDrawLine(graph_xgr2scr(oldx),graph_ygr2scr(oldy),
						graph_xgr2scr(x),graph_ygr2scr(y));
		}
		else if (graph_is_onscreen(x,y)) { 
			/* First point on screen */
			double newx,newy;
			graph_outscreen(&newx,&newy,x,y,oldx,oldy);
			WinDrawLine(graph_xgr2scr(x),graph_ygr2scr(y),
						graph_xgr2scr(newx),graph_ygr2scr(newy));
		}
		else if (graph_is_onscreen(oldx,oldy)) { 
			/* Second point on screen */
			double newx,newy;
			graph_outscreen(&newx,&newy,oldx,oldy,x,y);
			WinDrawLine(graph_xgr2scr(oldx),graph_ygr2scr(oldy),
						graph_xgr2scr(newx),graph_ygr2scr(newy));
		} 
	} 
	graphCurves[funcnum].x = x;
	graphCurves[funcnum].y = y;
	graphCurves[funcnum].valid = true;
}
/* Initialize cache variables to
 * avoid this time-expensive computation in the future */		
void
cacheTrans(void)
{		
	ScrPrefs.xtrans=(double)ScrPrefs.width/(graphPrefs.xmax-graphPrefs.xmin);
	ScrPrefs.ytrans=(double)ScrPrefs.height/(graphPrefs.ymax-graphPrefs.ymin);
}
		
/* Initializes Graph gadget and shows the basic grid */
static void 
InitGraphScreen(FormPtr frm,UInt16 graphId)
{
	UInt16 graphindex = FrmGetObjectIndex(frm,graphId);
	RectangleType bounds;
	Coord x,y;
	
	FrmGetObjectBounds(frm,graphindex,&bounds);
	WinEraseRectangle(&bounds,0);
	
	ScrPrefs.xmin=bounds.topLeft.x;
	ScrPrefs.width=bounds.extent.x;
	ScrPrefs.ymax=bounds.topLeft.y;
	ScrPrefs.height=bounds.extent.y;
	cacheTrans();

/* Now draw the x axis */
	x = graph_xgr2scr(0.0);
	y = graph_ygr2scr(0.0);
	if (x>bounds.topLeft.x && x<bounds.topLeft.x+bounds.extent.x)
	  WinDrawLine(x,
		      bounds.topLeft.y,
		      x,
		      bounds.topLeft.y+bounds.extent.y);
/* Now draw the y axis */
	if (y>bounds.topLeft.y && y<bounds.topLeft.y+bounds.extent.y)
	  WinDrawLine(bounds.topLeft.x,
		      y,
		      bounds.topLeft.x+bounds.extent.x,
		      y);

	
/* Draw the rectangle around the drawing area */
	WinDrawLine(bounds.topLeft.x-1,bounds.topLeft.y-1,bounds.topLeft.x-1,
		    bounds.topLeft.y+bounds.extent.y);
	WinDrawLine(bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y-1,bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y+bounds.extent.y);
	WinDrawLine(bounds.topLeft.x,bounds.topLeft.y-1,
		    bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y-1);
	WinDrawLine(bounds.topLeft.x-1,bounds.topLeft.y+bounds.extent.y,
		    bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y+bounds.extent.y);
}

static double
GetFnValue(Int16 funcnum,double x)
{
	return NaN;
}

/***********************************************************************
 *
 * FUNCTION:     graph_get_value
 * 
 * DESCRIPTION:  Return a real value of a given function with
 *               a given parameter
 *
 * PARAMETERS:   origstack - the compiled function
 *               param - argument of function
 *
 * RETURN:       result of function
 *               NaN - error, not defined etc.
 *      
 ***********************************************************************/
double
graph_get_value(CodeStack *origstack,double param)
{
	Trpn argument;
	CodeStack *stack;
	CError err;
	double result;
	
	stack = stack_new(origstack->size);
	stack_copy(stack,origstack);

	argument.type = real;
	argument.allocsize = 0;
	argument.u.realval = param;
	
	set_func_argument(&argument);
	err = stack_compute(stack);
	decr_func_argument();

	result = NaN;
	if (!err) {
		err = stack_get_val(stack,&result,real);
		if (err)
		  result = NaN;
	}
	stack_delete(stack);
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     graph_param_draw, graph_func_draw, graph_pol_draw
 * 
 * DESCRIPTION:  Draw a parametric/func/polar function
 *
 * PARAMETERS:   funcnum - index of a function to draw
 *               start - where to start drawing
 *               step - the interval, where to count points
 *               count - how many points to draw?
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_param_draw(Int16 funcnum,double start,double step,Int32 count)
{
	double x,y;
	Int32 i;
	
	for (i=0;i<count;i++) {
		x = graph_get_value(graphCurves[funcnum].stack1,start+i*step);
		y = graph_get_value(graphCurves[funcnum].stack2,start+i*step);
		graph_new_point(x,y,funcnum);
	}
}

static void
graph_func_draw(Int16 funcnum,double start,double step,Int32 count)
{       
	double x,y;
	Int32 i;

	for (i=0;i<count;i++) {
		x = start+i*step;
		y = graph_get_value(graphCurves[funcnum].stack1,x);
		graph_new_point(x,y,funcnum);
	}
}

static void
graph_polar_draw(Int16 funcnum,double start,double step,Int32 count)
{
	double x,y,len,angle;
	Int32 i;

	for (i=0;i<count;i++) {
		angle =  start+i*step;
		len = graph_get_value(graphCurves[funcnum].stack1,angle);
		if (finite(len)) {
			x = cos(angle) * len;
			y = sin(angle) * len;
		}
		else
		  x = y = NaN;
		graph_new_point(x,y,funcnum);
	}
}

static void
graph_zoom_out(FormPtr frm,Int16 objectId)
{
	double width = graphPrefs.xmax - graphPrefs.xmin;
	double height = graphPrefs.ymax - graphPrefs.ymin;
	
	graphPrefs.xmin-=width/2;
	graphPrefs.xmax+=width/2;
	
	graphPrefs.ymin-=height/2;
	graphPrefs.ymax+=height/2;
	
	FrmUpdateForm(tdFormGraph,frmUpdateVars);
}


/***********************************************************************
 *
 * FUNCTION:     graph_init_cache
 * 
 * DESCRIPTION:  Initializes the field of precompiled functions with
 *               NULLs
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_init_cache(void)
{
	Int16 i;
	for (i=0;i<MAX_GRFUNCS;i++) {
		graphCurves[i].stack1 = NULL;
		graphCurves[i].stack2 = NULL;
	}	  
}

/***********************************************************************
 *
 * FUNCTION:     graph_clean_cache
 * 
 * DESCRIPTION:  Frees memory occupied by precompiled functions
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_clean_cache(void)
{
	Int16 i;
	for (i=0;i<MAX_GRFUNCS;i++) {
		if (graphCurves[i].stack1) {
			stack_delete(graphCurves[i].stack1);
			graphCurves[i].stack1 = NULL;
		}
			if (graphCurves[i].stack2) {
			stack_delete(graphCurves[i].stack2);
			graphCurves[i].stack2 = NULL;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:     graph_compile_cache
 * 
 * DESCRIPTION:  Compiles functions that should be drawn and saves
 *               for later use
 *
 * PARAMETERS:   None
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_compile_cache(void)
{
	int i;
	char *name;
	CodeStack *stack;
	CError err;
	
	graph_clean_cache();
	if (graphPrefs.functype==graph_func ||
		graphPrefs.functype==graph_polar) {
		for (i=0;i<MAX_GRFUNCS;i++) {
			name = grsetup_get_fname(i);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack1 = stack;
		}
	} else {
		/* Parametric */
		for (i=0;i<MAX_GRFUNCS;i++) {
			name  = grsetup_get_fname(i*2);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack1 = stack;
			
			name  = grsetup_get_fname(i*2+1);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack2 = stack;
		}
	}		
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw_start
 * 
 * DESCRIPTION:  Initialize the drawing procedure, that later occurs 
 *               using nilEvent in background
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw_start(void)
{
	int i;
		
	for (i=0;i<MAX_GRFUNCS;i++) 
	  graphCurves[i].valid = false;
	
	graph_compile_cache();
	
	InitGraphScreen(FrmGetActiveForm(),graphGadget);
	
	graphState.active = true;
	graphState.position = 0;
	graphState.funcnum = 0;
	if (graphPrefs.functype == graph_func) {
		graphState.start = graphPrefs.xmin;
		graphState.stop = graphPrefs.xmax;
		graphState.step = (graphPrefs.xmax - graphPrefs.xmin)/ScrPrefs.width;
	} 
	else if (graphPrefs.functype == graph_polar) {
		graphState.start = graphPrefs.fimin;
		graphState.stop = graphPrefs.fimax;
		graphState.step = graphPrefs.fistep;
	}
	else if (graphPrefs.functype == graph_param) {
		graphState.start = graphPrefs.tmin;
		graphState.stop = graphPrefs.tmax;
		graphState.step = graphPrefs.tstep;
	}
	calc_nil_timeout(0);
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw_stop, graph_draw_resum
 * 
 * DESCRIPTION:  Stop/resume drawing functions
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw_resume()
{
	graphState.active = true;
	calc_nil_timeout(0);
}

static void
graph_draw_stop()
{
	if (graphState.active)
	  calc_nil_timeout(evtWaitForever);
	graphState.active = false;
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw
 * 
 * DESCRIPTION:  Draws a piece of a function, called from nilEvent
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw(void)
{
	Int32 i,scount;
	double start;

	if (!graphState.active)
	  return;
	if (graphState.funcnum>=MAX_GRFUNCS) {
		graph_draw_stop();
		return;
	}

	/* Check, if we have all functions to draw graph */
	if (!graphCurves[graphState.funcnum].stack1 ||
		(graphPrefs.functype==graph_param && !graphCurves[graphState.funcnum].stack2)) {
		graphState.funcnum++;
		return;
	}
	
	/* Should we draw all points at once, or step slowly to the end? */
	start = graphState.start + graphState.step * graphState.position;
	if (start + graphState.step * DRAW_AT_ONCE < graphState.stop)
	  scount = DRAW_AT_ONCE;
	else
	  scount = 1;
	
	for (i=0;i < DRAW_AT_ONCE && start <= graphState.stop;i+=scount) {
		if (graphPrefs.functype == graph_func) 
		  graph_func_draw(graphState.funcnum, start,graphState.step, scount);
		else if (graphPrefs.functype == graph_polar) 
		  graph_polar_draw(graphState.funcnum,start,graphState.step,scount);
		else if (graphPrefs.functype == graph_param)
		  graph_param_draw(graphState.funcnum,start,graphState.step,scount);
		start = graphState.start+graphState.step*(graphState.position+i);
	}
	
	graphState.position+=i;
	if (start >= graphState.stop) {
		/* Reached end of drawing area/parameters */
		graphState.position = 0;
		graphState.funcnum++;
	}
}

Boolean 
GraphFormHandleEvent(EventPtr event)
{
	Boolean    handled = false;
	FormPtr    frm=FrmGetActiveForm();
	Int16 controlId;

	switch (event->eType) {
	 case winExitEvent: /* This doesn't work, unfortunately */
		graph_draw_resume();
		break;
	 case nilEvent:
		graph_draw();
		handled = true;
		break;
	 case penDownEvent:		  
		handled=grtaps_tap(event->screenX,event->screenY);
		break;
	 case ctlSelectEvent:
		controlId=event->data.ctlSelect.controlID;
		  if (chooseForm(controlId)) {
			  handled=true;
			  break;
		  }
		handled=true;
		switch (controlId) {			
		 case graphZoomOut:
			graph_zoom_out(frm,graphGadget);
			break;
		 case graphPrefButton:
			FrmPopupForm(grPrefForm);
			break;
		 case graphSetupButton:
			FrmPopupForm(grSetupForm);
			break;
		 case graphTrack:
			if (CtlGetValue(GetObjectPtr(graphTrack)))
			  if (!grtaps_track_but())
				CtlSetValue(GetObjectPtr(graphTrack),false);
			graph_draw_resume();
			break;
		 default:
			handled=false;
		}
		break;
	 case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateVars
			|| event->data.frmUpdate.updateCode == frmRedrawUpdateCode)
		  graph_draw_start();
		if (event->data.frmUpdate.updateCode == frmUpdateVars)
		  handled = true;
		break;
	 case keyDownEvent:
		if (TxtCharIsHardKey(event->data.keyDown.modifiers,event->data.keyDown.chr)) {
			GotHardKey();
			handled=true;
		}
		break;
	 case frmCloseEvent:
		graph_draw_stop();
		graph_clean_cache();
		handled = false;
		break;
	 case frmOpenEvent:
		calcPrefs.form=frm->formId;		
		FrmDrawForm(frm);
		
		CtlSetValue(GetObjectPtr(tdBGra),1);
		graph_init_cache();
		graph_draw_start();
		handled=true;
		break;
	 case menuEvent:
		if (chooseForm(event->data.menu.itemID)) {
			handled=true;
			break;
		  }
		switch (event->data.menu.itemID) {
		 case graphPrefButton:
			FrmPopupForm(grPrefForm);
			handled = true;
			break;
		 case graphSetupButton:
			FrmPopupForm(grSetupForm);
			handled = true;
			break;
		 case tdBPref:
			FrmPopupForm(prefForm);
			handled = true;
			break;
		}
		break;
	}
	  
	return handled;
}
