/*
 *   $Id: grprefs.c,v 2.7 2000/09/28 19:29:11 ondra Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "grprefs.h"
#include "calcrsc.h"
#include "defuns.h"
#include "MathLib.h"
#include "stack.h"
#include "prefs.h"
#include "calc.h"
#include "display.h"
#include "mathem.h"

/* Local structure - values in function mode */
TgrPrefs graphPrefs;

/***********************************************************************
 *
 * FUNCTION:     grpref_init
 * 
 * DESCRIPTION:  Initialize graphical preferences, reading the appropriate
 *               preference from system
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
grpref_init(void)
{
	UInt16 size;
	Int16 i;
	
	size=sizeof(graphPrefs);
	if (PrefGetAppPreferences(APP_ID,PREF_GRAPH,&graphPrefs,&size,
							  true)!=PREF_GRAPH_VERSION) {
		graphPrefs.xmin = -7.0;
		graphPrefs.xmax = 7.0;
		graphPrefs.ymin = -7.0;
		graphPrefs.ymax = 7.0;
		graphPrefs.fimin = 0.0;
		graphPrefs.fimax = 2*M_PIl;
		graphPrefs.fistep = 0.08;
		graphPrefs.tmin = 0.0;
		graphPrefs.tmax = 10.0;
		graphPrefs.tstep = 0.1;
		graphPrefs.functype = graph_func;
		for (i=0;i<MAX_GRFUNCS;i++)
		  graphPrefs.funcFunc[i][0]='\0';
		for (i=0;i<MAX_GRFUNCS;i++)
		  graphPrefs.funcPol[i][0]='\0';
		for (i=0;i<MAX_GRFUNCS;i++) {
			graphPrefs.funcPar[i][0][0]='\0';
			graphPrefs.funcPar[i][1][0]='\0';
		}
	}	
}

/***********************************************************************
 *
 * FUNCTION:     grpref_close
 * 
 * DESCRIPTION:  Saves graph preferences
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
grpref_close(void)
{
	PrefSetAppPreferences(APP_ID,PREF_GRAPH,PREF_GRAPH_VERSION,&graphPrefs,
						  sizeof(graphPrefs),true);
}

/***********************************************************************
 *
 * FUNCTION:     grpref_compl_field
 * 
 * DESCRIPTION:  Computes the 'real' value of a text field
 *
 * PARAMETERS:   objectID - id of the field
 *
 * RETURN:       CError - 0 on success
 *               *value - the computed value
 *      
 ***********************************************************************/
static CError grpref_comp_field(Int16 objectID,double *value) IFACE;
static CError
grpref_comp_field(Int16 objectID,double *value)
{
	char *text;
	CError err;
	Tint_mode oldmode;
	CodeStack *stack;
	
	text=FldGetTextPtr(GetObjectPtr(objectID));
	
	oldmode=calcPrefs.int_mode;
	calcPrefs.int_mode=floating;
	
	stack = text_to_stack(text,&err);
	if (!err) {
		err = stack_compute(stack);
		if (!err)
		  err=stack_get_val(stack,(void *)value,real);
		stack_delete(stack);		  
	}
	calcPrefs.int_mode = oldmode;

	return err || !finite(*value);
}

/***********************************************************************
 *
 * FUNCTION:     grpref_save_values
 * 
 * DESCRIPTION:  Gets text from fields, computes them and saves them
 *               into corresponding global structure
 *
 * PARAMETERS:   None
 *
 * RETURN:       0 - success
 *               1 - error while computing
 *      
 ***********************************************************************/
static Int16 grpref_save_values() IFACE;
static Int16
grpref_save_values()
{
	double xmin,xmax,ymin,ymax;
	double t1,t2,t3;
	Tfunctype functype;	
	
	if (grpref_comp_field(grPrefXmin,&xmin))
	  return 0;
	if (grpref_comp_field(grPrefXmax,&xmax))
	  return 0;
	if (grpref_comp_field(grPrefYmin,&ymin))
	  return 0;
	if (grpref_comp_field(grPrefYmax,&ymax))
	  return 0;
	
	if (xmin>=xmax || ymin>=ymax)
	  return 0;
	
	if (CtlGetValue(GetObjectPtr(grPrefFunc)))
	  functype = graph_func;
	else {
		if (grpref_comp_field(grPrefField1,&t1))
		  return 0;
		if (grpref_comp_field(grPrefField2,&t2))
		  return 0;
		if (grpref_comp_field(grPrefField3,&t3))
		  return 0;
		if (CtlGetValue(GetObjectPtr(grPrefPol))) {
			if (t1>t2 || (t2-t1)/t3 > 10000) /* check that min<max and step is reasonable */
			  return 0;
			graphPrefs.fimin = t1;
			graphPrefs.fimax = t2;
			graphPrefs.fistep = t3;
			functype = graph_polar;
		}
		else {
			if (t1>t2 || (t2-t1)/t3 > 10000) /* check that min<max and step is reasonable */
			  return 0;
			graphPrefs.tmin = t1;
			graphPrefs.tmax = t2;
			graphPrefs.tstep = t3;
			functype = graph_param;
		}
	}	
	graphPrefs.functype = functype;
	graphPrefs.xmin=xmin;graphPrefs.xmax=xmax;
	graphPrefs.ymin=ymin;graphPrefs.ymax=ymax;
	return 1;
}

/***********************************************************************
 *
 * FUNCTION:     grpref_set_field
 * 
 * DESCRIPTION:  Sets a text in a field to a value
 *
 * PARAMETERS:   objectID - id of the field
 *               newval - value that should be printed into field
 *
 * RETURN:       
 *      
 ***********************************************************************/
static void grpref_set_field(UInt16 objectID,double newval) IFACE;
static void
grpref_set_field(UInt16 objectID,double newval)
{
	FieldPtr pole = GetObjectPtr(objectID);	  
	char *ptr;
	char *text;
	
	text = display_real(newval);
	
	ptr = FldGetTextPtr(pole);
	if (ptr)
	  FldDelete(pole,0,StrLen(ptr));
	FldInsert(pole,text,StrLen(text));
	MemPtrFree(text);
}

/***********************************************************************
 *
 * FUNCTION:     grpref_set_fields
 * 
 * DESCRIPTION:  Sets the basic xmin/xmax/ymin/ymax text values
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_set_fields(void) IFACE;
static void
grpref_set_fields(void)
{
	grpref_set_field(grPrefXmin,graphPrefs.xmin);
	grpref_set_field(grPrefXmax,graphPrefs.xmax);
	grpref_set_field(grPrefYmin,graphPrefs.ymin);
	grpref_set_field(grPrefYmax,graphPrefs.ymax);
}

/***********************************************************************
 *
 * FUNCTION:     grpref_show_fields
 * 
 * DESCRIPTION:  Shows additinal input fields, used by Polar/Parameter 
 *               functions
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_show_fields(void) IFACE;
static void
grpref_show_fields(void)
{
	FormPtr frm;
	
	frm = FrmGetActiveForm();
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefLabel1));
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefLabel2));
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefLabel3));
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefField1));
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefField2));
	FrmShowObject(frm,FrmGetObjectIndex(frm,grPrefField3));	
}

/***********************************************************************
 *
 * FUNCTION:     grpref_hide_fields
 * 
 * DESCRIPTION:  Hides additional fields used by polar/parametric graphs
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_hide_fields(void) IFACE;
static void
grpref_hide_fields(void)
{
	FormPtr frm;
	
	frm = FrmGetActiveForm();
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefLabel1));
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefLabel2));
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefLabel3));
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefField1));
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefField2));
	FrmHideObject(frm,FrmGetObjectIndex(frm,grPrefField3));
}

/***********************************************************************
 *
 * FUNCTION:     grpref_set_names
 * 
 * DESCRIPTION:  Set names of the 3 additional fields
 *
 * PARAMETERS:   txt1,2,3 - names of the fields
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_set_names(char *txt1,char *txt2,char *txt3) IFACE;
static void
grpref_set_names(char *txt1,char *txt2,char *txt3)
{
	FormPtr frm = FrmGetActiveForm();
	FrmCopyLabel(frm,grPrefLabel1,txt1);
	FrmCopyLabel(frm,grPrefLabel2,txt2);
	FrmCopyLabel(frm,grPrefLabel3,txt3);
}

/***********************************************************************
 *
 * FUNCTION:     grpref_pol_setup
 * 
 * DESCRIPTION:  Setup fields corresponding to polar graphs
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_pol_setup(void) IFACE;
static void
grpref_pol_setup(void)
{
	grpref_set_names("fi-min:","fi-max:","fi-step:");
	grpref_set_field(grPrefField1,graphPrefs.fimin);
	grpref_set_field(grPrefField2,graphPrefs.fimax);
	grpref_set_field(grPrefField3,graphPrefs.fistep);
	grpref_show_fields();
}

/***********************************************************************
 *
 * FUNCTION:     grpref_par_setup
 * 
 * DESCRIPTION:  Setup fields corresponding to parametric graphs
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grpref_par_setup(void) IFACE;
static void
grpref_par_setup(void)
{
	grpref_set_names("T-min:","T-max:","T-step:");
	grpref_set_field(grPrefField1,graphPrefs.tmin);
	grpref_set_field(grPrefField2,graphPrefs.tmax);
	grpref_set_field(grPrefField3,graphPrefs.tstep);
	grpref_show_fields();
}

/***********************************************************************
 *
 * FUNCTION:     GraphPrefsHandleEvent
 * 
 * DESCRIPTION:  Event handler for graph preferences
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean
GraphPrefsHandleEvent(EventPtr event)
{
	Boolean handled=false;
	Int16 controlID;
	FormPtr frm;
	Int16 formId;
	
	switch (event->eType) {
	 case frmOpenEvent:
		frm=FrmGetActiveForm();
		formId=frm->formId;
		FrmDrawForm(frm);
		grpref_set_fields();
		switch (graphPrefs.functype) {
		 default:
		 case graph_func:
			controlID = grPrefFunc;
			break;
		 case graph_polar:
			controlID = grPrefPol;
			grpref_pol_setup();
			break;
		 case graph_param:
			controlID = grPrefPar;
			grpref_par_setup();
			break;
		}
		CtlSetValue(GetObjectPtr(controlID),1);
		
		handled=true;
		break;
	 case ctlSelectEvent:
		controlID=event->data.ctlSelect.controlID;
		switch (controlID) {
		 case grPrefFunc:
		 case grPrefPol:
		 case grPrefPar:
			grpref_hide_fields();
			if (controlID==grPrefPol)
			  grpref_pol_setup();
			else if (controlID==grPrefPar)
			  grpref_par_setup();
			handled = true;
			break;
		 case grPrefDefault:
			grpref_set_field(grPrefXmin,-7.0);
			grpref_set_field(grPrefXmax,7.0);
			grpref_set_field(grPrefYmin,-7.0);
			grpref_set_field(grPrefYmax,7.0);
			handled=true;
			break;
		 case grPrefTrig:
			grpref_set_field(grPrefXmin,-2.0*M_PIl);
			grpref_set_field(grPrefXmax,2.0*M_PIl);
			grpref_set_field(grPrefYmin,-2.0*M_PIl);
			grpref_set_field(grPrefYmax,2.0*M_PIl);
			handled=true;
			break;			
		 case prefOK:
			if (grpref_save_values()) {
				FrmReturnToForm(calcPrefs.form);
				FrmUpdateForm(calcPrefs.form,frmUpdateVars);
			}
			else 
			  FrmAlert(tdGraphBadVal);
			handled=true;
			break;
		 case prefCancel:
			FrmReturnToForm(calcPrefs.form);
			handled=true;
			break;
		}
	}	
	  
	return handled;
}
