/*
 *   $Id: complex.c,v 2.9 2000/10/01 21:34:21 ondra Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "MathLib.h"
#include "konvert.h"
#include "stack.h"
#include "funcs.h"
#include "mathem.h"
#include "complex.h"
#include "defuns.h"
#include "prefs.h"

#define ALLOWED_ERROR 1E-15

/***********************************************************************
 *
 * FUNCTION:     cplx_to_gon
 * 
 * DESCRIPTION:  Convert an (a+bi) number to xe^(iy)
 *
 * PARAMETERS:   x - number in (a+bi) format
 *
 * RETURN:       number in r*e^(i*angle)
 *      
 ***********************************************************************/
Complex_gon
cplx_to_gon(Complex x)
{
	Complex_gon result;
	
	result.r = hypot(x.real,x.imag);
	if (result.r!=0.0) {
		if (x.real>=0.0)
		  result.angle = asin(x.imag/result.r);
		else if (x.imag>=0.0)		  
		  result.angle = M_PIl - asin(x.imag/result.r);
		else /* This isn't needed, but it looks better */
		  result.angle = -M_PIl - asin(x.imag/result.r);
	}
	else
	  result.angle=0.0;
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_multiply
 * 
 * DESCRIPTION:  Multiply complex numbers 
 *               (a+bi)*(c+di) = (a*c-b*d) + (a*d+b*c)i
 * PARAMETERS:   2 complex numbers
 *
 * RETURN:       result
 *      
 ***********************************************************************/
static Complex
cplx_multiply(Complex arg1,Complex arg2) MLIB;
static Complex
cplx_multiply(Complex arg1,Complex arg2)
{
	Complex result;
	
	result.real = arg1.real*arg2.real-arg1.imag*arg2.imag;
	result.imag = arg1.real*arg2.imag+arg1.imag*arg2.real;
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_power
 * 
 * DESCRIPTION:  base^expon computation, base & expon are complex
 *               z = xe^(iy)
 *               z^w = (xe^(iy))^w = x^w*e^(iyw) =
 *               = e^(w*ln(x))*e^(iyw) = w^(w*(ln(x)+yi))
 *
 * PARAMETERS:   2 complex numbers
 *
 * RETURN:       complex number
 *      
 ***********************************************************************/
static Complex
cplx_power(Complex base,Complex expon) MLIB;
static Complex
cplx_power(Complex base,Complex expon)
{
	Complex_gon gtmp;
	Complex tmp,result;
	
	gtmp=cplx_to_gon(base);
	tmp.real = log(gtmp.r);
	tmp.imag = gtmp.angle;
	
	tmp = cplx_multiply(expon,tmp);
	
	result.real = exp(tmp.real) * cos(tmp.imag);
	result.imag = exp(tmp.real) * sin(tmp.imag);
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_power_int
 * 
 * DESCRIPTION:  compute z^x, z is from C and x is integer
 *
 * PARAMETERS:   base, exponent
 *
 * RETURN:       result
 *      
 ***********************************************************************/
static Complex cplx_power_int(Complex_gon base,Int16 expon) MLIB;
static Complex
cplx_power_int(Complex_gon base,Int16 expon)
{
	Complex_gon res;
	Complex result;
	
	res.angle = base.angle * expon;
	
	/* Make it little faster then 'pow'*/
	res.r = 1.0;
	while (expon--)
	  res.r*=base.r;
	
	result.real = res.r * cos(res.angle);
	result.imag = res.r * sin(res.angle);
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     hlp_fact
 * 
 * DESCRIPTION:  simple n! computing
 *
 * PARAMETERS:   n 
 *
 * RETURN:       n!
 *      
 ***********************************************************************/
static double hlp_fact(Int16 arg) MLIB;
static double
hlp_fact(Int16 arg)
{
	double res=1.0;
	
	while (arg>0.0) {
		res*=arg;
		arg = arg - 1.0;
	}
	return res;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_add
 * 
 * DESCRIPTION:  the '+' with complex numbers
 *
 * PARAMETERS:   arg1, arg2
 *
 * RETURN:       arg1 + arg2
 *      
 ***********************************************************************/
static Complex cplx_add(Complex arg1,Complex arg2) MLIB;
static Complex
cplx_add(Complex arg1,Complex arg2)
{
	Complex result;
	
	result.real = arg1.real + arg2.real;
	result.imag = arg1.imag + arg2.imag;
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_cos
 * 
 * DESCRIPTION:  Cosinus in complex numbers, uses Taylors pol
 *
 * PARAMETERS:   arg
 *
 * RETURN:       cos(arg)
 *      
 ***********************************************************************/
static Complex cplx_cos(Complex arg) MLIB;
static Complex
cplx_cos(Complex arg)
{
	double error;
	Complex result,tmp;
	Complex_gon arggon;
	Int16 i;
	
	result.real = 1.0;
	result.imag = 0.0;
	
	arggon = cplx_to_gon(arg);
	
	error = ALLOWED_ERROR;
	for (i=2;i<100 && error>=ALLOWED_ERROR;i+=2) {
		tmp = cplx_power_int(arggon,i);
		tmp.real/=hlp_fact(i);
		tmp.imag/=hlp_fact(i);
		if (i%4) {
			tmp.real=-tmp.real;
			tmp.imag=-tmp.imag;
		}
		result = cplx_add(result,tmp);
		error = hypot(tmp.real,tmp.imag);
	}
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_sin
 * 
 * DESCRIPTION:  sinus in complex numbers, uses Taylors polynom
 *
 * PARAMETERS:   arg
 *
 * RETURN:       sin(arg)
 *      
 ***********************************************************************/
Complex cplx_sin(Complex arg) MLIB;
Complex
cplx_sin(Complex arg)
{
	double error;
	Complex result,tmp;
	Complex_gon arggon;
	Int16 i;
	
	result.real = arg.real;
	result.imag = arg.imag;
	
	arggon = cplx_to_gon(arg);
	
	error = ALLOWED_ERROR;
	for (i=3;i<100 && error>=ALLOWED_ERROR;i+=2) {
		tmp = cplx_power_int(arggon,(double)i);
		tmp.real/=hlp_fact(i);
		tmp.imag/=hlp_fact(i);
		if ((i-1)%4) {
			tmp.real=-tmp.real;
			tmp.imag=-tmp.imag;
		}
		result = cplx_add(result,tmp);
		error = hypot(tmp.real,tmp.imag);
	}
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_div
 * 
 * DESCRIPTION:  Division in complex numbers
 *              ! doesn't check for division by zero
 *
 * PARAMETERS:   arg1, arg2
 *
 * RETURN:       arg1/arg2
 *      
 ***********************************************************************/
static Complex
cplx_div(Complex arg1,Complex arg2) MLIB;
static Complex
cplx_div(Complex arg1,Complex arg2)
{
	Complex result;	
	double tmp;
	
	tmp = arg2.real*arg2.real + arg2.imag*arg2.imag;
	result.real = (arg1.real*arg2.real+arg1.imag*arg2.imag)/tmp;
	result.imag = (arg1.imag*arg2.real-arg1.real*arg2.imag)/tmp;
	
	return result;
}


/***********************************************************************
 *
 * FUNCTION:     cplx_oper
 *
 * DESCRIPTION:  Operations on complex numbers
 *
 * PARAMETERS:   On stack - 2 arguments convertible to complex
 *
 * RETURN:       On stack - 1 real number
 *      
 ***********************************************************************/
CError
cplx_oper(Int16 funcnum,CodeStack *stack)
{
	Complex arg1,arg2;
	Complex result;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,complex)))
	  return err;
	switch (funcnum) {
	 case FUNC_PLUS:
		result.real=arg1.real+arg2.real;
		result.imag=arg1.imag+arg2.imag;
		break;
	 case FUNC_MINUS:
		result.real=arg1.real-arg2.real;
		result.imag=arg1.imag-arg2.imag;
		break;
	 case FUNC_DIVIDE: 
		if (arg2.real==0.0 && arg2.imag==0.0)
		  return c_divbyzero;
		result = cplx_div(arg1,arg2);
		break;
	 case FUNC_MULTIPLY:
		result = cplx_multiply(arg1,arg2);
		break;
	 case FUNC_POWER:
		result = cplx_power(arg1,arg2);
		break;
	 case FUNC_MOD:
		if (arg1.imag!=0.0 || arg2.imag!=0.0)
		  return c_badarg;
		result.real = fmod(arg1.real,arg2.real);
		result.imag = 0.0;		
		break;
	 default:
		return c_badfunc;
	}       
	stack_add_val(stack,&result,complex);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     cplx_math
 * 
 * DESCRIPTION:  General mathematical functions on complex numbers
 *
 * PARAMETERS:   On stack - 1 number, convertible to complex
 *
 * RETURN:       On stack - 1 real or complex result
 *      
 ***********************************************************************/
CError
cplx_math(Int16 funcnum,CodeStack *stack)
{
	Complex carg,result;
	double res;
	CError err;
	
	err=stack_get_val(stack,&carg,complex);
		if (err)
	  return err;
	
	if (funcnum==MATH_ABS) {		
		res = hypot(carg.real,carg.imag);
		stack_add_val(stack,&res,real);		  
		return c_noerror;
	}
	if (funcnum==MATH_EXP) {
		result.real = exp(carg.real) * cos(carg.imag);
		result.imag = exp(carg.real) * sin(carg.imag);
		stack_add_val(stack,&result,complex);		
		return c_noerror;
	}
	if (funcnum==MATH_SQRT) {
		Complex_gon tmp;
		
		tmp=cplx_to_gon(carg);
		result.real = sqrt(tmp.r)*cos(tmp.angle/2);
		result.imag = sqrt(tmp.r)*sin(tmp.angle/2);
		stack_add_val(stack,&result,complex);
		return c_noerror;
	}
	if (funcnum==MATH_LN || funcnum==MATH_LOG || 
		funcnum==MATH_LOG2) {
		Complex_gon tmp;
		
		tmp = cplx_to_gon(carg);
		result.real = log(tmp.r);
		result.imag = tmp.angle;
		if (funcnum==MATH_LOG) {
			result.real/=log(10);
			result.imag/=log(10);
		}
		else if (funcnum==MATH_LOG2) {
			result.real/=log(2);
			result.imag/=log(2);
		}			
		stack_add_val(stack,&result,complex);
		return c_noerror;
	}
	if (funcnum==MATH_SIN || funcnum==MATH_COS ||
		funcnum==MATH_TAN) {
		
		if (calcPrefs.trigo_mode==degree) {
			carg.real=(2*M_PIl/360)*carg.real;
			carg.imag=(2*M_PIl/360)*carg.imag;
		}
		else if (calcPrefs.trigo_mode==grad) {
			carg.real=(2*M_PIl/400)*carg.real;
			carg.imag=(2*M_PIl/400)*carg.imag;
		}
		
		/* Fallback to faster computing, if number is not complex */
		if (carg.imag == 0.0) {
			double res;
			switch (funcnum) {
			 case MATH_SIN:
				res = sin(carg.real);
				break;
			 case MATH_COS:
				res = cos(carg.real);
				break;
			 case MATH_TAN:
			 default:
				res = tan(carg.real);
				break;
			}
			stack_add_val(stack,&res,real);
			return c_noerror;
		}
		switch (funcnum) {
		 case MATH_SIN:
			result = cplx_sin(carg);
			break;
		 case MATH_COS:
			result = cplx_cos(carg);
			break;
		 case MATH_TAN:
			result = cplx_div(cplx_sin(carg),cplx_cos(carg));
			break;
		}
		stack_add_val(stack,&result,complex);
		return c_noerror;
	}
	return c_internal;
}
