/*
 *   $Id: funcs.c,v 2.27 2000/09/28 23:12:34 ondra Exp $
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include <StringMgr.h>

#include "defuns.h"
#include "konvert.h"
#include "funcs.h"
#include "calcDB.h"
#include "MathLib.h"
#include "stack.h"
#include "calc.h"
#include "mathem.h"
#include "complex.h"
#include "history.h"
#include "prefs.h"

Trpn *func_argument;
Int16 func_argcount;

static CError
x_func(Trpn *rpn)
{
	if (!func_argcount)
	  return c_badarg;
	rpn_duplicate(rpn,*func_argument);
	return c_noerror;
}

CError 
eval_x(Int16 funcnum,CodeStack *stack)
{
	CError err;
	Trpn tmp;
	
	err = x_func(&tmp);
	if (err)
	  return err;
	stack_push(stack,tmp);
	return c_noerror;
}

static CError
pi_func(Trpn *rpn)
{
	rpn->type=real;
	rpn->u.realval=3.14159265358979311600L;
	rpn->allocsize = 0;
	return c_noerror;	
}

static CError
e_func(Trpn *rpn)
{
	rpn->type=real;
	rpn->u.realval=2.7182818284590452354L;
	rpn->allocsize = 0;
	return c_noerror;	
}

static CError
i_func(Trpn *rpn)
{
	rpn->type=complex;
	rpn->u.cplxval = MemPtrNew(sizeof(Complex));
	rpn->allocsize = sizeof(Complex);
	rpn->u.cplxval->real = 0.0;
	rpn->u.cplxval->imag = 1.0;
	return c_noerror;
}

struct {
	char *name;
	CError (*func)(Trpn *rpn);
}ConstTable[]={
	  {"pi",pi_func},
	  {"e",e_func},
	  {"i",i_func},
	  {NULL,NULL}
};

/* Take variable *source and put it's content to *dest */
/* Source & dest CAN be the same */
CError 
rpn_eval_variable(Trpn *dest,Trpn source)
{
	Int16 i;
	Trpn tmp;
	CError err;
	
	if (source.type!=variable) 	      
	  return c_internal;
	
/* Compare it with constants */	
	for (i=0;ConstTable[i].name;i++) 
	  if (StrCompare(ConstTable[i].name,source.u.varname)==0)
	    return ConstTable[i].func(dest);
	
	tmp = db_read_variable(source.u.varname,&err);
	if (err)
	  return err;
	
	*dest=tmp;
	
	return c_noerror;
}

CError
set_ans_var(Trpn ans)
{
	CError err;
	if (ans.type==variable) {
		err=rpn_eval_variable(&ans,ans);
		if (err)
		  return err;
	}
	db_write_variable("ans",ans);
	history_add(ans);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     change_param_name
 * 
 * DESCRIPTION:  Allows the user to change the name of the 'x' parameter
 *               to something else. I'm not sure if this function
 *               should be here..
 * 
 * NOTE:         Doesn't do any error checking, probably shouldn't be used
 *
 * PARAMETERS:   On stack - string, new name
 *
 * RETURN:       On stack - string, new name
 *      
 ***********************************************************************/
CError change_param_name(Int16 funcnum,CodeStack *stack)
{
	CError err;
	char *name;
	char *oldname;
	Trpn tmp;
	
	err = stack_get_val(stack,&name,string);
	if (err)
	  return err;
	
	if (StrLen(name)>MAX_FUNCNAME)
	  return c_badarg;
	oldname = MemPtrNew(StrLen(parameter_name)+1);
	StrCopy(oldname,parameter_name);
	StrCopy(parameter_name,name);
	/* rpn_get_val returns dynamically created variable */
	MemPtrFree(name);
		
    tmp.type=string;
	tmp.u.stringval=oldname;
	tmp.allocsize=StrLen(oldname)+1;
	stack_push(stack,tmp);
	
	return c_noerror;
}


/***********************************************************************
 *
 * FUNCTION:     unset_variable
 * 
 * DESCRIPTION:  Remove the variable/function
 *
 * PARAMETERS:   Variable/funcempty
 *
 * RETURN:       On stack - 0, always succeeds (event if it doesn't exist)
 *      
 ***********************************************************************/
CError
unset_variable(Int16 funcnum,CodeStack *stack)
{
	Trpn tmp;
	Int32 i = 0;
	
	tmp = stack_pop(stack);
		
	if (tmp.type==variable)
	  db_delete_record(tmp.u.varname);
	else if (tmp.type==funcempty)
	  db_delete_record(tmp.u.funcname);
	else {
		rpn_delete(tmp);
		return c_badarg;
	}
	rpn_delete(tmp);

	stack_add_val(stack,&i,integer);
	
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     define_func
 * 
 * DESCRIPTION:  Handler for 'a=something' and 'f()="something"'
 *               computes the value of something and saves the result
 *               in a database
 *
 * PARAMETERS:   On stack - variable/funcempty, value of the var/func
 *
 * RETURN:       funcempty or variable
 *      
 ***********************************************************************/
CError 
define_func(Int16 funcnum,CodeStack *stack)
{
	CError err;
	Int16 i;
	Trpn item1,item2;

	item1 = stack_pop(stack);
	item2 = stack_pop(stack);
	
	/* Definition of function */
	if (item2.type == funcempty && item1.type == string) {
		CodeStack *newstack;

		newstack=text_to_stack(item1.u.stringval,&err);
		if (err)
		  return err;
		db_write_function(item2.u.funcname,newstack,item1.u.stringval);

		stack_delete(newstack);

		rpn_delete(item1);
		item1.type=funcempty;
		StrCopy(item1.u.funcname,item2.u.funcname);
		item1.allocsize=0;
		stack_push(stack,item1);
		rpn_delete(item2);
		return c_noerror;
	}
	
	/* This is special, because stack_item_type evaluates variables */
	if (item2.type!=variable)
	  return c_badarg;
	
/* Check, if there isn't constant with same name */       
	for (i=0;ConstTable[i].name;i++)
	  if (StrCompare(item2.u.varname,ConstTable[i].name)==0)
	    return c_notvar;
	
	if (item1.type==variable) {
		err=rpn_eval_variable(&item1,item1);
		if (err)
		  return err;
	}
	db_write_variable(item2.u.varname,item1);
	rpn_delete(item1);
	stack_push(stack,item2);
	
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     text_function
 * 
 * DESCRIPTION:  Calls a user defined function (f())
 *
 * PARAMETERS:   name - name of the text function
 *               On stack - parameter for a function
 *
 * RETURN:       On stack - result
 *      
 ***********************************************************************/
CError
text_function(char *name,CodeStack *stack)
{
	Trpn *oldfuncarg;
	Trpn tmp;
	CError err;
	CodeStack *newstack;

/* The recursion should be less then MAX_RECURS */
	if (func_argcount>MAX_RECURS)
	  return c_deeprec;

	newstack = db_read_function(name,&err);
	if (err)
	  return err;

	tmp = stack_pop(stack);
	if (tmp.type==variable) {
		err=rpn_eval_variable(&tmp,tmp);
		if (err) {
			stack_delete(newstack);
			return err;
		}
	}
	/* Set 'tmp' as function argument */
	oldfuncarg=func_argument;
	func_argcount++;
	func_argument=&tmp;

	err = stack_compute(newstack);
	if (!err) 
	  stack_push(stack,stack_pop(newstack));

	/* Set function argument back to original */
	func_argcount--;
	func_argument=oldfuncarg;
	rpn_delete(tmp);

	stack_delete(newstack);
	return err;
}

/***********************************************************************
 *
 * FUNCTION:     empty_arg
 * 
 * DESCRIPTION:  Empty operator, that doesn't modify the stack
 *               Used mainly for the parameter delimiter (':')
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
CError
empty_arg(Int16 funcnum,CodeStack *stack)
{
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     discard_arg
 * 
 * DESCRIPTION:  Discards the argument, used for a ';' operator
 *
 * PARAMETERS:   On stack - 1 parameter
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
CError 
discard_arg(Int16 funcnum,CodeStack *stack)
{
	Trpn tmp;
	
	if (stack->size<1)
	  return c_noerror;

	tmp = stack_pop(stack);
	rpn_delete(tmp);

	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     set_func_argument
 * 
 * DESCRIPTION:  Push a parameter to a parameter stack
 * 
 * NOTE:         The parameter doesn't get copied, it can thus be
 *               later modified and the changes will take effect
 *               on every new stack_compute without the need
 *               of decr_func_argument. The argument gets copied
 *               upon calling the 'x' variable name 
 *
 * PARAMETERS:   arg - argument to be a new 'x'
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
set_func_argument(Trpn *arg)
{
	func_argument=arg;
	func_argcount++;
}

/***********************************************************************
 *
 * FUNCTION:     decr_func_argument
 * 
 * DESCRIPTION:  Remove a function parameter from a parameter stack
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
decr_func_argument(void)
{
	func_argcount--;
}

CError
convert_cplx_to_string(Int16 funcnum,CodeStack *stack)
{
	Complex arg;
	Complex_gon tmp;
	CError err;
	char *res;
	
	err=stack_get_val(stack,&arg,complex);
	if (err)
	  return err;

	if (funcnum == FUNC_GONIO) {
		res = MemPtrNew(MAX_FP_NUMBER*2+10);
	
		tmp = cplx_to_gon(arg);
		DoubleToAscii(tmp.r,res);
		StrCat(res,"e^(");
		DoubleToAscii(tmp.angle,res+StrLen(res));
		StrCat(res,"i)");
	}
	else if (funcnum == FUNC_TOCIS) {
		Complex_gon tmp = cplx_to_gon(arg);
		double angle;

		res = MemPtrNew(MAX_FP_NUMBER*2+20);
		angle = tmp.angle;
		if (calcPrefs.trigo_mode == degree) 
		  angle *= (360/(2*M_PIl));
		else if (calcPrefs.trigo_mode == grad) 
		  angle *= (400/(2*M_PIl));
		DoubleToAscii(tmp.r,res);
		StrCat(res,"(cos(");
		DoubleToAscii(angle,res+StrLen(res));
		StrCat(res,")+i*sin(");
		DoubleToAscii(angle,res+StrLen(res));
		StrCat(res,"))");
	}
	else
	  return c_internal;
	stack_add_val(stack,res,string);
	MemPtrFree(res);
	
	return c_noerror;
}

CError 
convert_real_to_string(Int16 funcnum,CodeStack *stack)
{
	double arg;
	char *res,*tmp;
	CError err;

	err=stack_get_val(stack,&arg,real);
	if (err)
	  return err;
	
	if (!finite(arg)) 
	  return c_badarg;
	
	res = MemPtrNew(MAX_FP_NUMBER+10);
	tmp = res;
	if (funcnum==FUNC_RADIAN) {
		DoubleToAscii(arg/M_PIl,res);
		StrCat(res,"pi");
	}	
	else if (funcnum==FUNC_DEGREE) {
		*tmp++='\225';      
		if (arg<0) {
			*tmp++='-';
			arg=-arg;
		}
		
		DoubleToAscii(trunc(arg),tmp);
		tmp+=StrLen(tmp);
		*tmp++='';
		if (arg-trunc(arg)<1E-15)
		  StrCopy(tmp,"0");
		else
		  DoubleToAscii(60*(arg-trunc(arg)),tmp);	
	}

	stack_add_val(stack,res,string);
	MemPtrFree(res);
	
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     varfunc_name_ok
 * 
 * DESCRIPTION:  Checks if the name of variable/function is correct
 *               in the system, and if it doesn't clash with some
 *               function name/constant name/parameter name
 *
 * PARAMETERS:   name - name to be checked
 *               type - function or variable
 * 
 * RETURN:       true - name ok
 *               false - name incorrect
 *      
 ***********************************************************************/
Boolean
varfunc_name_ok(char *name,rpntype type)
{
	Int16 i;
	
	if (StrLen(name)==0 || StrLen(name)>MAX_FUNCNAME)
	  return false;
	for (i=0;name[i];i++)
	  if (!IS_FUNC_LETTER(name[i]))
		return false;
	if (StrCompare(name,parameter_name)==0)
	  return false;
	if (type==variable) {
		for (i=0;ConstTable[i].name;i++)
		  if (StrCompare(name,ConstTable[i].name)==0)
			return false;
	}
	else if (type==function) {
		for (i=0;defined_funcs[i].name[0];i++)
		  if (StrCompare(name,defined_funcs[i].name)==0)
			return false;
	}
	return true;
}
