/*
 *   $Id: guess.c,v 2.9 2000/09/25 20:23:25 ondra Exp $
 * 
 *   Scientific Calculator for Palm's and PilmPilot's.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include <StringMgr.h>

#include "segment.h"
#include "MathLib.h"
#include "stack.h"
#include "funcs.h"
#include "konvert.h"
#include "complex.h"
#include "display.h"
#include "defuns.h"

#define MAX_DIV_F    1000
#define MAX_FORM_F 100
#define MAX_DIV_CONST 3
#define GUESS_MAX 100000


#define M_PI     3.14159265358979323846
#define M_E      2.7182818284590452354

static Boolean is_int(double t1) MLIB;
static Boolean
is_int(double t1)
{
	if (IS_ZERO(t1-round(t1)))
	  return true;
	return false;
}

static Boolean guess_div(double t1,Int16 *upper,Int16 *lower,Int16 maxlev) MLIB;
static Boolean
guess_div(double t1,Int16 *upper,Int16 *lower,Int16 maxlev)
{            
	Int16 i;
	double tmp;

	for (i=1;i<maxlev;i++) {
		tmp=i*t1;
		if (is_int(tmp)) {
			*upper=round(i*t1);
			*lower=i;
			return true;
		}
	}	
	return false;
}

static Boolean guess_mult_pi(double t1,Int16 *upper,Int16 *lower,Int16 *power) MLIB;
static Boolean
guess_mult_pi(double t1,Int16 *upper,Int16 *lower,Int16 *power)
{
	Int16 i;
	double tmp;
	
	/* First try multiplying it */
	tmp=t1;
	for (i=1;i<MAX_DIV_CONST;i++) {
		tmp*=M_PI;
		if (guess_div(tmp,upper,lower,MAX_FORM_F)) {
			*power=-i;
			return true;
		}
	}
	/* Dividing it */
	tmp=t1;
	for (i=1;i<MAX_DIV_CONST;i++) {
		tmp/=M_PI;
		if (guess_div(tmp,upper,lower,MAX_FORM_F)) {
			*power=i;
			return true;
		}
	}
	return false;
}

static Boolean guess_e_x(double t1,Int16 *upper,Int16 *lower) MLIB;
static Boolean
guess_e_x(double t1,Int16 *upper,Int16 *lower)
{
	if (t1<=0)
	  return false;	
	t1=log(t1);
	if (guess_div(t1,upper,lower,MAX_FORM_F))
	  return true;
	return false;
}

static Boolean guess_log_x(double t1,Int16 *upper,Int16 *lower) MLIB;
static Boolean
guess_log_x(double t1,Int16 *upper,Int16 *lower) 
{
	t1=exp(t1);
	if (guess_div(t1,upper,lower,MAX_FORM_F))
	  return true;
	return false;
}

static Boolean guess_pow1(double t1,Int16 *upper,Int16 *lower,Int16 *power) MLIB;
static Boolean
guess_pow1(double t1,Int16 *upper,Int16 *lower,Int16 *power)
{
	Int16 i;
	double tmp;
	
	if (t1<=0.0)
	  t1=-t1;
	
	tmp=t1;
	for (i=2;i<MAX_DIV_CONST;i++) {
		tmp*=t1;
		if (guess_div(tmp,upper,lower,MAX_FORM_F)) {
			*power=i;
			return 1;
		}
	}
	return 0;
}

static void 
convert_pow1(char *result,Int16 negative,Int16 upper,Int16 lower,Int16 power) MLIB;
static void
convert_pow1(char *result,Int16 negative,Int16 upper,Int16 lower,Int16 power)
{
	if (negative) {
		StrCopy(result,"-");
		result++;
	}
	if (power==2 && lower==1)
	  StrPrintF(result,"sqrt(%d)",upper);
	else if (power==2)
	  StrPrintF(result,"sqrt(%d/%d)",upper,lower);
	else if (lower==1) 
	  StrPrintF(result,"%d^(1/%d)",upper,power);
	else 
	  StrPrintF(result,"(%d/%d)^(1/%d)",upper,lower,power);
}

static void 
convert_pi(char *result,Int16 upper,Int16 lower,Int16 power) MLIB;
static void 
convert_pi(char *result,Int16 upper,Int16 lower,Int16 power)
{
	result[0]='\0';
	if (upper<0) {
		upper=-upper;
		StrCopy(result,"-");
		result++;		
	}	  
	if (power>0) {
		if (upper>1) 
		  StrPrintF(result,"%d*",upper);
		result+=StrLen(result);
		if (power==1)
		  StrPrintF(result,"pi");
		else
		  StrPrintF(result,"pi^%d",power);
		result+=StrLen(result);
		if (lower>1)
		  StrPrintF(result,"/%d",lower);		
	}
	else if (power<0) {
		StrPrintF(result,"%d/",upper);
		result+=StrLen(result);
		if (lower>1)
		  StrPrintF(result,"(%d*",lower);
		StrCat(result,"pi");
		result+=StrLen(result);
		if (power<-1)
		  StrPrintF(result,"^%d",-power);
		if (lower>1)
		  StrCat(result,")");
	}	   	  
}


static Boolean guess_real(double t1,char *result) MLIB;
static Boolean
guess_real(double t1,char *result)
{
	Int16 upper,lower,power;
	
	if (!finite(t1))
	  return false;

	if (fabs(t1)>100000)
	  return false;
	
	if (is_int(t1)) {
		StrPrintF(result,"%d",(int)round(t1));
		return true;
	}      
	if (guess_div(t1,&upper,&lower,MAX_DIV_F)) {
		StrPrintF(result,"%d/%d",upper,lower);
		return true;
	}
	if (guess_e_x(t1,&upper,&lower)) {
		if (lower==1 && upper>0) 
		  StrPrintF(result,"e^%d",upper);
		else if (lower==1)
		  StrPrintF(result,"e^(%d)",upper);
		else
		  StrPrintF(result,"e^(%d/%d)",upper,lower);
		return true;
	}
	if (guess_mult_pi(t1,&upper,&lower,&power)) {
		convert_pi(result,upper,lower,power);
		return true;		
	}       
	if (guess_log_x(t1,&upper,&lower)) {
		if (lower==1)
		  StrPrintF(result,"ln(%d)",upper);
		else
		  StrPrintF(result,"ln(%d/%d)",upper,lower);
		return true;
	}
	if (guess_pow1(t1,&upper,&lower,&power)) {
		if (upper && lower) {
			convert_pow1(result,t1<0.0,upper,lower,power);
			return true;
		}
	}
	return 0;	
}

static Boolean
guess_complex(Complex arg,char *result) MLIB;
static Boolean
guess_complex(Complex arg,char *result)
{
	Complex_gon arg_gon;
	char *text;
	Int16 upper,lower,power;
	Boolean found=false;
	double tmp;
		
	if (!finite(arg.real) || !finite(arg.imag))
	  return false;

	if (fabs(arg.real)>100000 || fabs(arg.imag)>100000)
	  return false;
	
	arg_gon = cplx_to_gon(arg);
	
	if (is_int(arg_gon.r)) 
	  StrPrintF(result,"%d",(int)round(arg_gon.r));
	else if (guess_div(arg_gon.r,&upper,&lower,MAX_DIV_F)) {
		StrPrintF(result,"(%d/%d)",upper,lower);
		found = true;
	}
	else if (guess_pow1(arg_gon.r,&upper,&lower,&power) &&
			 upper && lower) {
		convert_pow1(result,arg_gon.r<0.0,upper,lower,power);
	}
	else {
		text = display_real(arg_gon.r);
		StrCopy(result,text);
		MemPtrFree(text);
	}
	
	tmp = arg_gon.angle/M_PI;	
	if (is_int(tmp) && (int)round(tmp)!=0) {
		StrPrintF(result+StrLen(result),"e^(%dpi*i)",(int)round(tmp));
		found = true;
	}
	else if (guess_div(tmp,&upper,&lower,MAX_DIV_F) && upper) {
		found = true;
		StrCat(result,"e^(");
		if (upper<0) {
			StrCat(result,"-");
			upper = -upper;
		}
		if (upper==1)
		  StrPrintF(result+StrLen(result),"pi*i/%d",lower);
		else
		  StrPrintF(result+StrLen(result),"%dpi*i/%d",upper,lower);
		StrCat(result,")");
	}
	return found;
}

char * guess(Trpn item) MLIB;
char *
guess(Trpn item)
{
	char *result;
	Boolean guessed;
	
	if (item.type == real) {
		result = MemPtrNew(MAX_FP_NUMBER+1);
		guessed = guess_real(item.u.realval,result);
		if (!guessed) {
			MemPtrFree(result);
			return NULL;
		}
		return result;
	}
	else if (item.type == complex) {
		result = MemPtrNew(MAX_FP_NUMBER*3);
		guessed = guess_complex(*item.u.cplxval,result);
		if (!guessed && IS_ZERO(item.u.cplxval->imag))
		  guessed = guess_real(item.u.cplxval->real,result);
		if (!guessed) {				
			MemPtrFree(result);
			return NULL;
		}
		return result;
	}
	return NULL;
}
