/*
 *   $Id: history.c,v 2.5 2000/09/25 23:34:54 ondra Exp $
 * 
 *   Scientific Calculator for Palm's and PilmPilot's.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "calcDB.h"
#include "stack.h"
#include "defuns.h"
#include "history.h"
#include "display.h"

static DmOpenRef dbRef;

UInt16
history_total(void) MLIB;
UInt16
history_total(void)
{
	return DmNumRecords(dbRef);
}

/***********************************************************************
 *
 * FUNCTION:     history_add
 * 
 * DESCRIPTION:  Add a record in a records database
 *
 * PARAMETERS:   item - the record that should be added
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
history_add(Trpn item)
{
	dbStackItem *packItem,*newItem;
	UInt16 size;
	MemHandle newHandle;
	UInt16 recordNumber;
	
	packItem = rpn_pack_record(item);
	size = sizeof(*packItem) + packItem->datasize;
	
	recordNumber = 0;
	newHandle = DmNewRecord(dbRef,&recordNumber,size);	
	newItem = MemHandleLock(newHandle);
	DmWrite(newItem,0,packItem,size);
	MemHandleUnlock(newHandle);
	DmReleaseRecord(dbRef,recordNumber,true);
	
	MemPtrFree(packItem);
}

/***********************************************************************
 *
 * FUNCTION:     history_shrink
 * 
 * DESCRIPTION:  Shrink the history detabase to a defined
 *               number of records
 *
 * PARAMETERS:   How many records leave in database
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void 
history_shrink(UInt16 count)
{
	UInt16 i;

	if (DmNumRecords(dbRef)==0)
	  return;
	i = DmNumRecords(dbRef)-1;
	while (i>count) {
		DmRemoveRecord(dbRef,i--);
	}
}


/***********************************************************************
 *
 * FUNCTION:     history_get
 * 
 * DESCRIPTION:  Return a history record
 *
 * PARAMETERS:   num - how old the item is
 *
 * RETURN:       item
 *      
 ***********************************************************************/
Trpn history_get(UInt16 num) MLIB;
Trpn 
history_get(UInt16 num)
{
	MemHandle recordHandle;
	dbStackItem *packItem;
	Trpn result;
	
	recordHandle = DmQueryRecord(dbRef,num);
	packItem = MemHandleLock(recordHandle);
	result = rpn_unpack_record(packItem);
	MemHandleUnlock(recordHandle);
	
	return result;
}
	

/***********************************************************************
 *
 * FUNCTION:     history_command
 * 
 * DESCRIPTION:  This function is evaluated directly from a normal
 *               mathematical equation
 *
 * PARAMETERS:   On stack - 1 number convertible to integer
 *
 * RETURN:       On stack - value from history database
 *      
 ***********************************************************************/
CError
history_command(Int16 funcnum,CodeStack *stack)
{
	CError err;
	UInt32 arg;
	
	if ((err=stack_get_val(stack,&arg,integer)))
	  return err;	
	if (arg > DmNumRecords(dbRef))
	  return c_badarg;

	stack_push(stack,history_get(arg));
	
	return c_noerror;
}

static Err history_open_byid(LocalID dbid) MLIB;
static Err
history_open_byid(LocalID dbid)
{
        dbRef=DmOpenDatabase(CARDNO,dbid,dmModeReadWrite);
        if (dbRef)
          return DmGetLastErr();                        
        return 0;
}

/***********************************************************************
 *
 * FUNCTION:     history_open
 * 
 * DESCRIPTION:  Open a history database
 *
 * PARAMETERS:   None
 *
 * RETURN:       0 on success
 *      
 ***********************************************************************/
Int16 
history_open(void)
{
	Err err;
	LocalID dbid;
	UInt16 version;
	UInt32 creator;
	UInt32 type;
	
	dbid=DmFindDatabase(CARDNO,HISTORYDBNAME);
	if (dbid) {  /* Database exists */
		DmDatabaseInfo(CARDNO,dbid,
			       NULL, /* name */
                               NULL, /* attrib */
                               &version, /* version */
                               NULL, /* crDate */
                               NULL, /* modDate */
                               NULL, /* bckUpDate */
                               NULL, /* modNum */
                               NULL, /* appinfoID */
                               NULL, /* sortInfoID */
                               &type, /* Type */
                               &creator); /* Creator */

		if (version==DBVERSION && creator==LIB_ID && type==DBTYPE)
		  return history_open_byid(dbid);
		
		/* Database exists, but with uncorrect version */
		err=DmDeleteDatabase(CARDNO,dbid);
		if (err)
		  return err;
	}
	
	/* Database doesn't exist or old version */
	err=DmCreateDatabase(CARDNO,HISTORYDBNAME,LIB_ID,DBTYPE,false);
	if (err)
	  return err;
	
	dbid=DmFindDatabase(CARDNO,HISTORYDBNAME);
	
	version=DBVERSION;
	DmSetDatabaseInfo(CARDNO,dbid,
			  NULL, /* name */
			  NULL, /* attrib */
			  &version, /* version */
			  NULL, /* crDate */
			  NULL, /* modDate */
			  NULL, /* bckUpDate */
			  NULL, /* modNum */
			  NULL, /* appinfoID */
			  NULL, /* sortInfoID */
			  NULL, /* Type */
			  NULL); /* Creator */
	return history_open_byid(dbid);
}

/***********************************************************************
 *
 * FUNCTION:     history_close
 *
 * DESCRIPTION:  Close a history database
 *
 * PARAMETERS:   None
 *
 * RETURN:       0 on success
 *      
 ***********************************************************************/
Int16 
history_close(void)
{
	history_shrink(HISTORY_RECORDS);
	return DmCloseDatabase(dbRef);
}
