/*   
 * $Id: varmgr.c,v 2.15 2000/09/26 16:44:37 ondra Exp $
 * 
 * Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include "konvert.h"
#include "calcDB.h"
#include "display.h"
#include "varmgr.h"
#include "calcrsc.h"
#include "stack.h"
#include "calc.h"
#include "funcs.h"
#include "history.h"

static char varEditTitle[MAX_EDIT_TITLE+1];
static char varEditName[MAX_FUNCNAME+1];
static rpntype varEditType;
static void (*varEditCallback)(void);
static Boolean varEditEditName;

static char HORIZ_ELLIPSIS;

static void hist_draw_item(Int16 itemNum,RectangleType *bounds,
						   Char **itemText) IFACE;
static void hist_draw_item(Int16 itemNum,RectangleType *bounds,
						   Char **itemText)
{
	Trpn item;
	char *text;
	Int16 len;
	
	item = history_get(itemNum);
	text = display_default(item);
	rpn_delete(item);

	len = StrLen(text);
	if (FntCharsWidth(text,len)>bounds->extent.x) {
		while (FntCharsWidth(text,len)>bounds->extent.x)
		  len--;
		text[len-1]=HORIZ_ELLIPSIS;
	}
	WinDrawChars(text,len,bounds->topLeft.x,bounds->topLeft.y);
	MemPtrFree(text);
}

/***********************************************************************
 *
 * FUNCTION:     history_popup
 * 
 * DESCRIPTION:  Displays a history list and returns a selected item
 *
 * PARAMETERS:   None
 *
 * RETURN:       history(n), where n is the item number
 *      
 ***********************************************************************/
char *
history_popup(void)
{
	ListPtr list;
	Int16 i;
	char *result;
	UInt16 numitems;
	
	numitems = history_total();
	if (numitems==0) 
	  return NULL;

	ChrHorizEllipsis(&HORIZ_ELLIPSIS);
	list = GetObjectPtr(histList);
	LstMakeItemVisible(list,0);
	LstSetListChoices(list,NULL,numitems);
	LstSetDrawFunction(list,hist_draw_item);
	LstSetHeight(list,
				 numitems>11?11:numitems);
	LstSetSelection(list,0);
	i = LstPopupList(list);
	
	if (i==noListSelection)
	  result = NULL;
	else {
		result = MemPtrNew(15); /* history(xxxx) */
		StrPrintF(result,"history(%d)",i);
	}
	return result;
}


/***********************************************************************
 *
 * FUNCTION:     varmgr_popup
 * 
 * DESCRIPTION:  Popups a menu with list of variables/functions
 *               and their values
 * 
 * PARAMETERS:   Type of what to display
 *
 * RETURN:       Name of selected variable/function or NULL
 *      
 ***********************************************************************/
char *
varmgr_popup(rpntype type)
{
	dbList *varlist;
	char **values;
	Trpn item;
	char *text;
	CError err;
	Int16 i;
	ListPtr list;
	
	varlist = db_get_list(type);
	if (varlist->size==0) {
		db_delete_list(varlist);
		return NULL;
	}
	values = MemPtrNew(varlist->size * sizeof(*values));
	for (i=0;i<varlist->size;i++) {
		if (type==variable) {
			item = db_read_variable(varlist->list[i],&err);
			if (!err) {
				text = display_default(item);
				rpn_delete(item);
			}
			else
			  text = print_error(err);
		}
		else {
			err = db_func_description(varlist->list[i],&text);
			if (err)
			  text = print_error(err);
		}
		values[i] = MemPtrNew(StrLen(varlist->list[i])+StrLen(text)+10);
		StrCopy(values[i],varlist->list[i]);
		if (type==function)
		  StrCat(values[i],"()");
		StrCat(values[i]," ->  ");
		StrCat(values[i],text);
		MemPtrFree(text);
	}
	
	list = GetObjectPtr(varList);
	LstSetListChoices(list,values,varlist->size);
	LstSetHeight(list,
				 varlist->size>MAX_LIST_LENGTH?MAX_LIST_LENGTH:varlist->size);
	list->attr.search = true; /* We have it sorted alphabetically */
	i = LstPopupList(list);
	if (i==noListSelection)
	  text = NULL;
	else {
		text = MemPtrNew(StrLen(varlist->list[i])+1);
		StrCopy(text,varlist->list[i]);
	}
	  	
	for (i=0;i<varlist->size;i++)
	  MemPtrFree(values[i]);
	MemPtrFree(values);
	db_delete_list(varlist);
	
	return text;
}

/***********************************************************************
 *
 * FUNCTION:     varmgr_save_variable
 * 
 * DESCRIPTION:  Called by an OK button in varEntryForm,
 *               checks if the variable has a correct name and saves
 *               the 'ans' result into it
 *
 * PARAMETERS:   None
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void varmgr_save_variable(void) IFACE;
static void
varmgr_save_variable()
{
	char *varname;
	Trpn item;
	CError err;
	
	varname = FldGetTextPtr(GetObjectPtr(varEntryField));
	if (!varname || !StrLen(varname)) {
		FrmAlert(alertVarProblem);
		return;
	}
	if (!varfunc_name_ok(varname,variable))
	  return;
	
	item = db_read_variable("ans",&err);
	if (err) {
		FrmAlert(alertAnsProblem);
		return;
	}
	db_write_variable(varname,item);
	rpn_delete(item);
	
	FrmReturnToForm(0);
}

/***********************************************************************
 *
 * FUNCTION:     varmgr_get_var
 * 
 * DESCRIPTION:  displays a small popup menu with names of variables
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static char * varmgr_get_var(void) IFACE;
static char *
varmgr_get_var(void)
{
	dbList *varlist;
	ListPtr list;
	Int16 i;
	char *result;
	
	varlist = db_get_list(variable);
	if (varlist->size==0) {
		db_delete_list(varlist);
		return NULL;
	}
	list = GetObjectPtr(varEntryList);
	LstSetListChoices(list,varlist->list,varlist->size);
	LstSetHeight(list,
				 varlist->size>5?5:varlist->size);
	list->attr.search=true;
	i = LstPopupList(list);
	if (i==noListSelection)
	  result = NULL;
	else {
		result = MemPtrNew(StrLen(varlist->list[i])+1);
		StrCopy(result,varlist->list[i]);
	}
	db_delete_list(varlist);
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     VarmgrEntryHandleEvent
 * 
 * DESCRIPTION:  Event handler for varEntryForm, used to enter
 *               a name of variable to save a 'ans'
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean
VarmgrEntryHandleEvent(EventPtr event)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 controlID;

	switch (event->eType) {
	 case keyDownEvent:
		if (event->data.keyDown.chr=='\n') {
			varmgr_save_variable();
			handled=true;
		}
		break;
	 case ctlSelectEvent:
		controlID = event->data.ctlSelect.controlID;
		if (controlID == varEntryCancel) {
			FrmReturnToForm(0);
			handled = true;
			break;
		}
		if (controlID == varEntryOK) {
			varmgr_save_variable();
			handled = true;
			break;
		}
		if (controlID == varEntryListBut) {
			char *varname;
			FieldPtr field;
			
			handled = true;
			
			varname = varmgr_get_var();
			if (!varname)
			  break;
			field = GetObjectPtr(varEntryField);
			if (FldGetTextPtr(field))
			  FldDelete(field,0,StrLen(FldGetTextPtr(field)));
			FldInsert(field,varname,StrLen(varname));
			MemPtrFree(varname);
			break;
		}
		break;
	 case frmOpenEvent:
		frm = FrmGetActiveForm();
		FrmDrawForm(frm);
		
		FrmSetFocus(frm,FrmGetObjectIndex(frm,varEntryField));
		
		handled = true;
		break;		
	 case frmCloseEvent:
		/* FrmHandleEvent must handle this event itself */
		handled = false;
		break;
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:     varmgr_edit_init
 * 
 * DESCRIPTION:  Initialize a form for Input of a value of a variable
 *
 * PARAMETERS:   Nothing, (globals verEditName,Description & type)
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void varmgr_edit_init(void) IFACE;
static void
varmgr_edit_init(void)
{
	FormPtr frm;
	char *text;
	char *fmtvalue;
	Trpn item;
	CError err;
	MemHandle namehandle;
	FieldPtr field;
	FieldAttrType fieldattr;
	
	frm = FrmGetActiveForm();

	FrmSetTitle(frm,varEditTitle);
	
	field = GetObjectPtr(varEditNameField);
	namehandle = MemHandleNew(MAX_FUNCNAME+1);
	text = MemHandleLock(namehandle);
	StrCopy(text,varEditName);
	MemHandleUnlock(namehandle);
	
	FldGetAttributes(field,&fieldattr);
	if (varEditEditName) {
		fieldattr.editable = true;
		fieldattr.underlined = grayUnderline;
	}
	else {
		fieldattr.editable = false;
		fieldattr.underlined = noUnderline;
	}
	FldSetAttributes(field,&fieldattr);
	FldSetTextHandle(field,namehandle);
		
	if (varEditType == variable) {
		item = db_read_variable(varEditName,&err);
		if (!err) {
			fmtvalue = display_default(item);
			FldInsert(GetObjectPtr(varEditField),fmtvalue,StrLen(fmtvalue));
			FldSetSelection(GetObjectPtr(varEditField),0,StrLen(fmtvalue));
			MemPtrFree(fmtvalue);
			rpn_delete(item);
		}	
	} 
	else if (varEditType == function) {
		err = db_func_description(varEditName,&fmtvalue);
		if (!err) {
			FldInsert(GetObjectPtr(varEditField),fmtvalue,StrLen(fmtvalue));
			FldSetSelection(GetObjectPtr(varEditField),0,StrLen(fmtvalue));
			MemPtrFree(fmtvalue);
		}
	}
}
/***********************************************************************
 *
 * FUNCTION:     varmgr_edit_save
 * 
 * DESCRIPTION:  Recomputes and saves value written into input field
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       true on success
 *      
 ***********************************************************************/
static Boolean varmgr_edit_save() IFACE;
static Boolean
varmgr_edit_save()
{
	CodeStack *stack;
	FieldPtr fldPtr;
	CError err;
	Trpn result;
	char *text;

	if (varEditEditName) {
		fldPtr = GetObjectPtr(varEditNameField);
		text = FldGetTextPtr(fldPtr);		
		if (!text || !varfunc_name_ok(text,varEditType)) {
			FrmAlert(varBadVariableName);
			return false;
		}
		if (db_record_exists(text) &&			
			FrmCustomAlert(varConfirmOverwrite,text,NULL,NULL)) 
		  return false;
		StrCopy(varEditName,text);
	}
		
	fldPtr = GetObjectPtr(varEditField);
	text = FldGetTextPtr(fldPtr);
	if (!text) {
		FrmAlert(tdAlertCompute);
		return false;
	}
	
	stack = text_to_stack(text,&err);
	
	if (err) {
		FrmAlert(tdAlertCompute);
		return false;
	}
	if (varEditType == variable) {
		err = stack_compute(stack);
		if (err) {
			FrmAlert(tdAlertCompute);
			stack_delete(stack);
			return false;
		}
		result = stack_pop(stack);
		stack_delete(stack);
	
		if (result.type == variable) {
			err = rpn_eval_variable(&result,result);
			if (err) {
				rpn_delete(result);
				FrmAlert(tdAlertCompute);
				return false;
			}
		}
		db_write_variable(varEditName,result);
		rpn_delete(result);
		return true;
	}
	else if (varEditType==function) {
		db_write_function(varEditName,stack,text);
		stack_delete(stack);
		return true;
	}
	return false;
}

/***********************************************************************
 *
 * FUNCTION:     VarmgrEditHandleEvent
 * 
 * DESCRIPTION:  Event handler for the variable Entry
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean
VarmgrEditHandleEvent(EventPtr event)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 controlID;

	switch (event->eType) {
	 case keyDownEvent:
		if (event->data.keyDown.chr=='\n') {
			if (varmgr_edit_save()) {
				if (varEditCallback)
				  varEditCallback();
				FrmReturnToForm(0);
			}
			handled=true;
		}
		break;
	 case ctlSelectEvent:
		controlID = event->data.ctlSelect.controlID;
		if (controlID == varEditCancel) {
			FrmReturnToForm(0);
			handled = true;
			break;
		}
		if (controlID == varEditOK) {
			if (varmgr_edit_save()) {
				if (varEditCallback)
				  varEditCallback();
				FrmReturnToForm(0);
			}
			handled = true;
			break;
		}
		break;
	 case frmOpenEvent:
		frm = FrmGetActiveForm();
		varmgr_edit_init();
		FrmDrawForm(frm);
		
		if (varEditEditName)
		  FrmSetFocus(frm,FrmGetObjectIndex(frm,varEditNameField));
		else
		  FrmSetFocus(frm,FrmGetObjectIndex(frm,varEditField));
		
		handled = true;
		break;		
	 case frmCloseEvent:
		/* FrmHandleEvent must handle this event itself */
		handled = false;
		break;
	}
	
	return handled;
}

/***********************************************************************
 *
 * FUNCTION:     varmgr_edit
 *
 * DESCRIPTION:  Popup a variable edit dialog
 *
 * PARAMETERS:   varname - name of the variable
 *               title - title of the new form
 *               type - function or variable
 *               callback - function, that should be called
 *                          upon succesful save (usually update)
 *               editname - user can edit the variable name and can
 *                          and can create a new variable/function
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void 
varmgr_edit(char *varname,char *title,rpntype type,void (*callback)(void),
			Boolean editname)
{
	StrCopy(varEditName,varname);
	StrCopy(varEditTitle,title);
	varEditType = type;
	varEditCallback = callback;
	varEditEditName = editname;
	FrmPopupForm(varEditForm);
}
